#!/usr/bin/env python
import os
import sys
import django
from django.conf import settings
from django.test.utils import get_runner


def run():
    verbosity = 1
    if "-v" in sys.argv or "--verbose" in sys.argv:
        verbosity = 2
    os.environ['DJANGO_SETTINGS_MODULE'] = 'ls.joyous.tests.settings'
    django.setup()
    TestRunner = get_runner(settings)
    test_runner = TestRunner(top_level="ls/joyous",
                             verbosity=verbosity,
                             keepdb=False)
    labels = ["ls.joyous.tests."+arg for arg in sys.argv[1:]
              if not arg.startswith("-")]
    if not labels:
        labels = ["ls.joyous.tests"]
    failures = test_runner.run_tests(labels)
    return failures

def coverage():
    from coverage import Coverage
    cover = Coverage()
    cover.start()
    failures = run()
    cover.stop()
    cover.save()
    cover.html_report()
    return failures

if __name__ == "__main__":
    if "--coverage" in sys.argv:
        failures = coverage()
    else:
        failures = run()
    sys.exit(bool(failures))



# ------------------------------------------------------------------------------
# Test Calendar Pages
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from unittest.mock import Mock
from django_bs_test import TestCase
from django.contrib.auth.models import User
from django.test import RequestFactory
from django.utils import translation
from django.urls import reverse
from wagtail.core.models import Site, Page
from ls.joyous.models.calendar import (CalendarPage, SpecificCalendarPage,
                                       GeneralCalendarPage)
from ls.joyous.models.events import SimpleEventPage
from ls.joyous.models.groups import get_group_model
from .testutils import freeze_timetz, getPage

GroupPage = get_group_model()

# ------------------------------------------------------------------------------
class TestCalendar(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@j.test', 's3(r3t')
        calendar = CalendarPage(owner  = self.user,
                                slug  = "events",
                                title = "Events",
                                default_view = "L")
        Page.objects.get(slug='home').add_child(instance=calendar)
        calendar.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "tree-planting",
                                title = "Tree Planting",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30),
                                time_to   = dt.time(11,0))
        calendar.add_child(instance=event)
        event.save_revision().publish()

    def testMonthView(self):
        response = self.client.get("/events/2012/03/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select('thead a')), 4)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "March")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 31)
        self.assertEqual(len(select("tbody td.noday")), 4)
        holidayNames = [holiday.string.strip() for holiday in
                        select("tbody td.day .holiday .holiday-name")]
        self.assertEqual(holidayNames,
                         ["Taranaki Anniversary Day", "Otago Anniversary Day"])

    def testMonthAbbr(self):
        response = self.client.get("/events/2012/Apr/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "April")

    def testWeekView(self):
        response = self.client.get("/events/2012/W11/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select('thead a')), 4)
        self.assertEqual(len(select("table.calendar thead tr th.sun")), 1)
        self.assertEqual(len(select("tbody tr")), 1)
        self.assertEqual(len(select("tbody td.day")), 7)
        holidays = select("tbody td.day .holiday")
        self.assertEqual(len(holidays), 1)
        self.assertEqual(holidays[0].h4.string.strip(), "12 Mar")
        self.assertEqual(holidays[0].div.string.strip(),
                         "Taranaki Anniversary Day")

    def testDayWithOneEvent(self):
        response = self.client.get("/events/2011/6/5/")
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/events/tree-planting/")

    def testDayWithEvents(self):
        calendar = getPage("/home/events/")
        event1 = SimpleEventPage(owner = self.user,
                                 slug  = "memory",
                                 title = "We remember",
                                 date      = dt.date(2011,6,8),
                                 time_from = dt.time(9,30),
                                 time_to   = dt.time(10,30))
        calendar.add_child(instance=event1)
        event1.save_revision().publish()
        event2 = SimpleEventPage(owner = self.user,
                                slug  = "action",
                                title = "Make things better",
                                date      = dt.date(2011,6,8),
                                time_from = dt.time(11,30),
                                time_to   = dt.time(16,0))
        calendar.add_child(instance=event2)
        event2.save_revision().publish()
        response = self.client.get("/events/2011/6/8/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".events-on-day")), 1)
        events = select(".events-on-day .event-item")
        self.assertEqual(len(events), 2)
        title = events[0].select("a.event-title")[0]
        self.assertEqual(title.string.strip(), "We remember")
        self.assertEqual(title['href'], "/events/memory/")

    def testUpcomingEvents(self):
        response = self.client.get("/events/upcoming/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".upcoming-events")), 1)
        self.assertEqual(len(select(".upcoming-events .event-item")), 0)

    def testUpcomingEventsInvalidPage(self):
        response = self.client.get("/events/upcoming/?page=99")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".upcoming-events")), 1)
        self.assertEqual(len(select(".upcoming-events .event-item")), 0)

    def testPastEvents(self):
        response = self.client.get("/events/past/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".past-events")), 1)
        events = select(".past-events .event-item")
        self.assertEqual(len(events), 1)
        title = events[0].select("a.event-title")[0]
        self.assertEqual(title.string.strip(), "Tree Planting")
        self.assertEqual(title['href'], "/events/tree-planting/")
        when = events[0].select(".event-when")[0]
        self.assertEqual(when.string.strip(),
                         "Sunday 5th of June 2011 at 9:30am to 11am")

    def testPastEventsInvalidPage(self):
        response = self.client.get("/events/past/?page=99")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".past-events")), 1)
        self.assertEqual(len(select(".past-events .event-item")), 1)

    def testRouteDefault(self):
        response = self.client.get("/events/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select(".upcoming-events")), 1)
        response = self.client.get("/events/?view=weekly")
        self.assertEqual(response.status_code, 200)
        select = response.soup.select
        self.assertEqual(len(select(".week-name")), 1)
        response = self.client.get("/events/?view=monthly")
        select = response.soup.select
        self.assertEqual(len(select(".month-name")), 1)

    def testMiniMonthView(self):
        response = self.client.get("/events/mini/2011/06/")
        self.assertEqual(response.status_code, 404)

    def testMiniMonthAjaxView(self):
        response = self.client.get("/events/mini/2011/06/",
                                   HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(select("table.minicalendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "June")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 30)
        self.assertEqual(len(select("tbody td.noday")), 5)
        event = select("tbody td.day a.event")[0]
        self.assertEqual(event.string.strip(), "5")
        self.assertEqual(event['href'], "/events/2011/06/05/")
        self.assertEqual(event['title'], "Tree Planting")

    def testInvalidDates(self):
        invalidDates = ["2012/13", "2008/W54", "2099/W53"]
        for date in invalidDates:
            with self.subTest(date=date):
                response = self.client.get("/events/{}/".format(date))
                self.assertEqual(response.status_code, 404)

    def testCalendarStartMonth(self):
        response = self.client.get("/events/1900/1/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 2)
        self.assertEqual(links[0].get('title'), "Next month")
        self.assertEqual(links[0].get('href'), "/events/1900/2/")
        self.assertEqual(links[1].get('title'), "Next year")
        self.assertEqual(links[1].get('href'), "/events/1901/1/")

    def testCalendarFinishMonth(self):
        response = self.client.get("/events/2099/12/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 2)
        self.assertEqual(links[0].get('title'), "Previous month")
        self.assertEqual(links[0].get('href'), "/events/2099/11/")
        self.assertEqual(links[1].get('title'), "Previous year")
        self.assertEqual(links[1].get('href'), "/events/2098/12/")

    def testCalendarStartWeek(self):
        response = self.client.get("/events/1900/W1/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links0 = select('.calendar-options .events-view a')
        self.assertEqual(len(links0), 3)
        self.assertEqual(links0[0].get_text(), "List View")
        self.assertEqual(links0[1].get_text(), "This Week")
        self.assertEqual(links0[2].get_text(), "Monthly View")
        self.assertEqual(links0[2].get('href'), "/events/1900/1/")
        links1 = select('thead a')
        self.assertEqual(len(links1), 2)
        self.assertEqual(links1[0].get('title'), "Next week")
        self.assertEqual(links1[0].get('href'), "/events/1900/W2/")
        self.assertEqual(links1[1].get('title'), "Next year")
        self.assertEqual(links1[1].get('href'), "/events/1901/W1/")

    def testCalendarFinishWeek(self):
        response = self.client.get("/events/2099/W52/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links0 = select('.calendar-options .events-view a')
        self.assertEqual(len(links0), 3)
        self.assertEqual(links0[0].get_text(), "List View")
        self.assertEqual(links0[1].get_text(), "This Week")
        self.assertEqual(links0[2].get_text(), "Monthly View")
        self.assertEqual(links0[2].get('href'), "/events/2099/12/")
        links1 = select('thead a')
        self.assertEqual(len(links1), 2)
        self.assertEqual(links1[0].get('title'), "Previous week")
        self.assertEqual(links1[0].get('href'), "/events/2099/W51/")
        self.assertEqual(links1[1].get('title'), "Previous year")
        self.assertEqual(links1[1].get('href'), "/events/2098/W52/")

    def testNextYearW53(self):
        response = self.client.get("/events/2098/W53/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        links = select('thead a')
        self.assertEqual(len(links), 3)
        self.assertEqual(links[0].get('title'), "Previous week")
        self.assertEqual(links[0].get('href'), "/events/2098/W52/")
        self.assertEqual(links[1].get('title'), "Next week")
        self.assertEqual(links[1].get('href'), "/events/2099/W1/")
        self.assertEqual(links[2].get('title'), "Previous year")
        self.assertEqual(links[2].get('href'), "/events/2097/W52/")

    def testFutureDates(self):
        response = self.client.get("/events/2525/1/")
        self.assertEqual(response.status_code, 404)
        response = self.client.get("/events/2100/W1/")
        self.assertEqual(response.status_code, 404)


# ------------------------------------------------------------------------------
class TestFrançais(TestCase):
    def setUp(self):
        translation.activate('fr')
        self.user = User.objects.create_superuser('i', 'i@j.test', 's3(r3t')
        calendar = CalendarPage(owner  = self.user,
                                slug  = "calendrier",
                                title = "Calendrier")
        Page.objects.get(slug='home').add_child(instance=calendar)
        calendar.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "plantation-d-arbres",
                                title = "Plantation d'arbres",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30),
                                time_to   = dt.time(11,0))
        calendar.add_child(instance=event)
        event.save_revision().publish()

    def tearDown(self):
        translation.deactivate()

    def testMonthView(self):
        response = self.client.get("/calendrier/2012/03/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "Mars")

    def testWeekView(self):
        response = self.client.get("/calendrier/2012/W11/")
        select = response.soup.select
        self.assertEqual(response.status_code, 200)
        holidays = select("tbody td.day .holiday")
        self.assertEqual(holidays[0].h4.string.strip(), "12 Mar")
        self.assertEqual(holidays[0].div.string.strip(),
                         "Taranaki Anniversary Day")


# ------------------------------------------------------------------------------
class TestSpecificCalendar(TestCase):
    def setUp(self):
        SpecificCalendarPage.is_creatable = True
        self.user = User.objects.create_user('i', 'i@j.test', 's3(r3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar1 = SpecificCalendarPage(owner  = self.user,
                                              slug  = "calendar1",
                                              title = "Red Team Calendar")
        home = getPage("/home/")
        home.add_child(instance=self.calendar1)
        self.calendar1.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "football",
                                title = "Football Game",
                                uid   = "570ed9c4-4503-4b45-b15e-c99faed9c531",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30),
                                time_to   = dt.time(11,0))
        self.calendar1.add_child(instance=event)
        event.save_revision().publish()
        self.calendar2 = SpecificCalendarPage(owner  = self.user,
                                              slug  = "calendar2",
                                              title = "Green Team Calendar")
        home.add_child(instance=self.calendar2)
        self.calendar2.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "tree-planting",
                                title = "Tree Planting",
                                uid   = "eb50e787-12bf-477b-8493-c4414ac001ca",
                                date      = dt.date(2011,6,5),
                                time_from = dt.time(9,30))
        self.calendar2.add_child(instance=event)
        event.save_revision().publish()

    def testEventsOnDay(self):
        evod = self.calendar1._getEventsOnDay(self.request, dt.date(2011,6,5))
        self.assertEqual(evod.date, dt.date(2011,6,5))
        self.assertEqual(len(evod.all_events), 1)
        self.assertEqual(len(evod.days_events), 1)
        event = evod.days_events[0]
        self.assertEqual(event.title, "Football Game")

    def testEventsByDay(self):
        evods = self.calendar1._getEventsByDay(self.request, dt.date(2011,7,1),
                                               dt.date(2011,7,10))
        self.assertEqual(len(evods), 10)
        evod = evods[0]
        self.assertEqual(evod.date, dt.date(2011,7,1))
        self.assertEqual(len(evod.all_events), 0)

    def testEventsByWeek(self):
        weeks = self.calendar1._getEventsByWeek(self.request, 2011, 6)
        self.assertEqual(len(weeks), 5)
        self.assertIsNone(weeks[0][0])
        self.assertIsNone(weeks[0][2])
        evod = weeks[1][0]
        self.assertEqual(evod.date, dt.date(2011,6,5))
        self.assertEqual(len(evod.all_events), 1)
        event = evod.days_events[0]
        self.assertEqual(event.title, "Football Game")
        self.assertEqual(event.page.slug, "football")

    @freeze_timetz("2011-01-21 15:00")
    def testUpcomingEvents(self):
        events = self.calendar2._getUpcomingEvents(self.request)
        self.assertEqual(len(events), 1)
        title, event, url = events[0]
        self.assertEqual(title, "Tree Planting")
        self.assertEqual(event.slug, "tree-planting")

    @freeze_timetz("2011-01-21 15:00")
    def testPastEvents(self):
        events = self.calendar2._getPastEvents(self.request)
        self.assertEqual(len(events), 0)

    def testGetEventFromUid(self):
        event = self.calendar1._getEventFromUid(self.request,
                                                "570ed9c4-4503-4b45-b15e-c99faed9c531")
        self.assertEqual(event.title, "Football Game")
        event = self.calendar1._getEventFromUid(self.request,
                                                "eb50e787-12bf-477b-8493-c4414ac001ca")
        self.assertIsNone(event)

    def testGetAllEventsByDay(self):
        events = self.calendar2._getAllEvents(self.request)
        self.assertEqual(len(events), 1)
        self.assertEqual(events[0].title, "Tree Planting")


# ------------------------------------------------------------------------------
class TestGeneralCalendar(TestCase):
    def setUp(self):
        GeneralCalendarPage.is_creatable = True
        self.user = User.objects.create_user('i', 'i@j.test', 's3(r3t')
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.calendar = GeneralCalendarPage(owner  = self.user,
                                             slug  = "calendar",
                                             title = "My Calendar")
        home = getPage("/home/")
        home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "bbq",
                                title = "BBQ",
                                uid   = "a94c3211-08e5-4e36-9448-86a869a47d89",
                                date      = dt.date(2011,8,20),
                                time_from = dt.time(18,30))
        self.calendar.add_child(instance=event)
        event.save_revision().publish()
        self.group = GroupPage(slug = "initech", title = "Initech Corporation")
        home.add_child(instance=self.group)
        self.group.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "planning-meeting",
                                title = "Planning to Plan",
                                uid   = "a96484da-e350-45c7-af03-97ca2bb173d6",
                                date      = dt.date(2011,8,20),
                                time_from = dt.time(8,0))
        self.group.add_child(instance=event)
        event.save_revision().publish()

    def testEventsOnDay(self):
        evod = self.calendar._getEventsOnDay(self.request, dt.date(2011,8,20))
        self.assertEqual(evod.date, dt.date(2011,8,20))
        self.assertEqual(len(evod.all_events), 2)
        self.assertEqual(len(evod.days_events), 2)
        events = evod.days_events
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[1].title, "BBQ")

    def testEventsByDay(self):
        evods = self.calendar._getEventsByDay(self.request, dt.date(2011,8,1),
                                              dt.date(2011,8,20))
        self.assertEqual(len(evods), 20)
        evod = evods[19]
        self.assertEqual(evod.date, dt.date(2011,8,20))
        self.assertEqual(len(evod.all_events), 2)
        events = evod.days_events
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[1].title, "BBQ")

    def testEventsByWeek(self):
        weeks = self.calendar._getEventsByWeek(self.request, 2011, 8)
        self.assertEqual(len(weeks), 5)
        self.assertIsNone(weeks[0][0])
        evod = weeks[2][6]
        self.assertEqual(evod.date, dt.date(2011,8,20))
        self.assertEqual(len(evod.all_events), 2)
        events = evod.days_events
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[1].title, "BBQ")

    @freeze_timetz("2011-08-21 15:00")
    def testUpcomingEvents(self):
        events = self.calendar._getUpcomingEvents(self.request)
        self.assertEqual(len(events), 0)

    @freeze_timetz("2011-08-21 15:00")
    def testPastEvents(self):
        events = self.calendar._getPastEvents(self.request)
        self.assertEqual(len(events), 2)
        title, event, url = events[0]
        self.assertEqual(title, "BBQ")
        self.assertEqual(event.slug, "bbq")
        title, event, url = events[1]
        self.assertEqual(event.slug, "planning-meeting")

    def testGetEventFromUid(self):
        event = self.calendar._getEventFromUid(self.request,
                                                "a94c3211-08e5-4e36-9448-86a869a47d89")
        self.assertEqual(event.title, "BBQ")
        event = self.calendar._getEventFromUid(self.request,
                                                "a96484da-e350-45c7-af03-97ca2bb173d6")
        self.assertEqual(event.title, "Planning to Plan")

    def testGetAllEventsByDay(self):
        events = self.calendar._getAllEvents(self.request)
        self.assertEqual(len(events), 2)
        self.assertEqual(events[0].title, "Planning to Plan")
        self.assertEqual(events[1].title, "BBQ")


# ------------------------------------------------------------------------------
class TestMultiCalendarCreate(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@j.test', 's3(r3t')
        self.main = getPage("/home/")
        self.sub = Page(slug="nova", title="Nova Homepage")
        self.main.add_child(instance=self.sub)
        self.sub.save_revision().publish()
        Site.objects.create(hostname='nova.joy.test',
                            root_page_id=self.sub.id,
                            is_default_site=False)
        SpecificCalendarPage.is_creatable = True
        GeneralCalendarPage.is_creatable = True

    def testMainSiteAnotherCalendar(self):
        calendar = CalendarPage(owner  = self.user,
                                slug  = "events",
                                title = "Events",)
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertFalse(CalendarPage.can_create_at(self.main))

    def testSubSiteAnotherCalendar(self):
        calendar = CalendarPage(owner  = self.user,
                                slug  = "events",
                                title = "Events")
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertTrue(CalendarPage.can_create_at(self.sub))

    def testNoSiteAnotherCalendar(self):
        rogue = Page(slug="rogue", title="Rogue")
        self.assertFalse(CalendarPage.can_create_at(rogue))

    def testMainSiteAnotherSpecificCalendar(self):
        calendar = SpecificCalendarPage(owner  = self.user,
                                        slug  = "events",
                                        title = "Events")
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertTrue(SpecificCalendarPage.can_create_at(self.main))

    def testMainSiteAnotherGeneralCalendar(self):
        calendar = GeneralCalendarPage(owner  = self.user,
                                       slug  = "events",
                                       title = "Events",)
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertFalse(GeneralCalendarPage.can_create_at(self.main))

    def testSubSiteAnotherGeneralCalendar(self):
        calendar = GeneralCalendarPage(owner  = self.user,
                                       slug  = "events",
                                       title = "Events",)
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertFalse(GeneralCalendarPage.can_create_at(self.sub))

    def testCalendarMixture(self):
        general = GeneralCalendarPage(owner  = self.user,
                                      slug  = "events1",
                                      title = "Events",)
        self.main.add_child(instance=general)
        general.save_revision().publish()
        self.assertTrue(CalendarPage.can_create_at(self.main))
        calendar = CalendarPage(owner  = self.user,
                                slug  = "events2",
                                title = "Events",)
        self.main.add_child(instance=calendar)
        calendar.save_revision().publish()
        self.assertTrue(SpecificCalendarPage.can_create_at(self.main))
        specific = SpecificCalendarPage(owner  = self.user,
                                        slug  = "events3",
                                        title = "Events")
        self.main.add_child(instance=specific)
        specific.save_revision().publish()

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Multiday Recurring Event Page
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
import calendar
from django.test import TestCase, RequestFactory
from django.contrib.auth.models import User
from django.utils import timezone
from wagtail.core.models import Page, PageViewRestriction
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import DAILY, WEEKLY, MONTHLY, YEARLY
from ls.joyous.utils.recurrence import SA, TU, TH, FR
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models.events import MultidayRecurringEventPage, ExtraInfoPage
from .testutils import datetimetz, freeze_timetz

# ------------------------------------------------------------------------------
class Test(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        Page.objects.get(slug='home').add_child(instance=self.calendar)
        self.calendar.save()
        self.calendar.save_revision().publish()
        self.event = MultidayRecurringEventPage(
                               owner = self.user,
                               slug  = "team-retreat",
                               title = "Team Retreat",
                               repeat = Recurrence(dtstart=dt.date(2000,1,1),
                                                   freq=YEARLY,
                                                   bymonth=8,
                                                   byweekday=FR(1)),
                               num_days  = 3,
                               time_from = dt.time(18),
                               time_to   = dt.time(16,30))
        self.calendar.add_child(instance=self.event)
        self.event.save_revision().publish()

    def testGetEventsByDay(self):
        events = MultidayRecurringEventPage.events.byDay(dt.date(2017,8,1),
                                                         dt.date(2017,8,31))
        self.assertEqual(len(events), 31)
        evod = events[3]
        self.assertEqual(evod.date, dt.date(2017,8,4))
        self.assertEqual(len(evod.days_events), 1)
        self.assertEqual(len(evod.continuing_events), 0)
        evod = events[4]
        self.assertEqual(evod.date, dt.date(2017,8,5))
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 1)
        evod = events[5]
        self.assertEqual(evod.date, dt.date(2017,8,6))
        self.assertEqual(len(evod.days_events), 0)
        self.assertEqual(len(evod.continuing_events), 1)

    def testStatus(self):
        with freeze_timetz("2014-08-01 17:00:00"):
            self.assertEqual(self.event.status_text, "")
        with freeze_timetz("2014-08-02 13:00:00"):
            self.assertEqual(self.event.status_text, "This event has started.")
        with freeze_timetz("2014-08-03 15:00:00"):
            self.assertEqual(self.event.status_text, "This event has started.")
        with freeze_timetz("2014-08-03 17:00:00"):
            self.assertEqual(self.event.status_text, "")

    def testWhen(self):
        self.assertEqual(self.event.when,
                         "The first Friday of August for 3 days "
                         "starting at 6pm finishing at 4:30pm")

    def testAt(self):
        self.assertEqual(self.event.at.strip(), "6pm")

    @freeze_timetz("2035-04-03 10:00:00")
    def testPrevDate(self):
        self.assertEqual(self.event.prev_date, dt.date(2034, 8, 4))

    @freeze_timetz("2018-04-03 10:00:00")
    def testFutureExceptions(self):
        request = RequestFactory().get("/test")
        request.user = self.user
        request.session = {}
        info2018 = ExtraInfoPage(owner = self.user,
                                 slug  = "2018-08-03-extra-info",
                                 title = "Extra-Info for Friday 3rd of August",
                                 overrides = self.event,
                                 except_date = dt.date(2018, 8, 3),
                                 extra_title = "Team Retreat 2018",
                                 extra_information = "Weekend at Bernie's")
        self.event.add_child(instance=info2018)
        exceptions = self.event._futureExceptions(request)
        self.assertEqual(len(exceptions), 1)
        info = exceptions[0]
        self.assertEqual(info.slug, "2018-08-03-extra-info")
        self.assertEqual(info.extra_title, "Team Retreat 2018")


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# Generated by Django 2.0.8 on 2018-09-20 00:10

from django.db import migrations, models
import ls.joyous.fields


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0009_alter_uid_notnull'),
    ]

    operations = [
        migrations.AddField(
            model_name='calendarpage',
            name='default_view',
            field=models.CharField(choices=[('L', 'List View'), ('W', 'Weekly View'), ('M', 'Monthly View')], default='M', max_length=15),
        ),
        migrations.AddField(
            model_name='calendarpage',
            name='view_choices',
            field=ls.joyous.fields.MultipleSelectField(blank=True, choices=[('L', 'List View'), ('W', 'Weekly View'), ('M', 'Monthly View')], default=['L', 'W', 'M']),
        ),
    ]



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# Autodoc may need to import some models modules which require django settings
# be configured
os.environ['DJANGO_SETTINGS_MODULE'] = 'ls.joyous.tests.settings'
import django
django.setup()

# Use SQLite3 database engine so it doesn't attempt to use psycopg2 on RTD
os.environ['DATABASE_ENGINE'] = 'django.db.backends.sqlite3'


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'


# -- Project information -----------------------------------------------------

project = 'Joyous'
copyright = '2019, Linuxsoftware'
author = 'Linuxsoftware'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

from ls.joyous import __version__

# The short X.Y version
version = '.'.join(__version__.split('.', 2)[:2])

# The full version, including alpha/beta/rc tags
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'

# on_rtd is whether we are on readthedocs.org, this line of code grabbed from docs.readthedocs.org
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Joyousdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Joyous.tex', 'Joyous Documentation',
     'Linuxsoftware', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'joyous', 'Joyous Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Joyous', 'Joyous Documentation',
     author, 'Joyous', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# -- Options for autodoc extension -------------------------------------------

autoclass_content = "both"
_yes = None # None is equivalent to giving the option name in the list format
autodoc_default_options = {'private-members':   _yes,}
#   'members':           _yes,
#   'undoc-members':     _yes,
#   'private-members':   _yes,
#   'special-members':   _yes,
#   'inherited-members': _yes,
#   'show-inheritance':  _yes, 

# -- Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    'python':   ('https://docs.python.org/3', None),
    'django':   ('http://docs.djangoproject.com/en/stable/',
                 'http://docs.djangoproject.com/en/stable/_objects/'),
    'wagtail':  ('https://docs.wagtail.io/en/stable', None),
    'dateutil': ('https://dateutil.readthedocs.io/en/stable', None),
    }

def setup(app):
#    app.add_stylesheet('css/custom.css')

    # Add the Django setting construct
    app.add_crossref_type(
        directivename = "setting",
        rolename = "setting",
        indextemplate = "pair: %s; setting",
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2018-02-09 02:57
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
import django.db.models.deletion
import ls.joyous.fields
import wagtail.contrib.routable_page.models
import wagtail.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
        ('wagtailimages', '0019_delete_filter'),
    ]

    operations = [
        migrations.CreateModel(
            name='CalendarPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('intro', wagtail.core.fields.RichTextField(blank=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(wagtail.contrib.routable_page.models.RoutablePageMixin, 'wagtailcore.page'),
        ),
        migrations.CreateModel(
            name='CancellationPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('except_date', models.DateField(help_text='For this date', verbose_name='For Date')),
                ('cancellation_title', models.CharField(blank=True, help_text='Show in place of cancelled event (Leave empty to show nothing)', max_length=255, verbose_name='Title')),
                ('cancellation_details', wagtail.core.fields.RichTextField(blank=True, help_text='Why was the event cancelled?', verbose_name='Details')),
            ],
            options={
                'verbose_name': 'Cancellation',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='EventCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=4, unique=True)),
                ('name', models.CharField(max_length=80)),
            ],
            options={
                'verbose_name_plural': 'Event Categories',
                'verbose_name': 'Event Category',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ExtraInfoPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('except_date', models.DateField(help_text='For this date', verbose_name='For Date')),
                ('extra_information', wagtail.core.fields.RichTextField(help_text='Information just for this date')),
            ],
            options={
                'verbose_name': 'Extra Event Information',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='GroupPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('content', wagtail.core.fields.RichTextField(blank=True, default='')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='MultidayEventPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('time_from', models.TimeField(blank=True, null=True, verbose_name='Start time')),
                ('time_to', models.TimeField(blank=True, null=True, verbose_name='End time')),
                ('location', models.CharField(blank=True, max_length=255)),
                ('details', wagtail.core.fields.RichTextField(blank=True)),
                ('website', models.URLField(blank=True)),
                ('date_from', models.DateField(default=datetime.date.today, verbose_name='Start date')),
                ('date_to', models.DateField(default=datetime.date.today, verbose_name='End date')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.EventCategory', verbose_name='Category')),
                ('group_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='multidayeventpage_events', related_query_name='multidayeventpage_event', to='joyous.GroupPage')),
                ('image', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailimages.Image')),
            ],
            options={
                'verbose_name': 'Multiday Event Page',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='RecurringEventPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('time_from', models.TimeField(blank=True, null=True, verbose_name='Start time')),
                ('time_to', models.TimeField(blank=True, null=True, verbose_name='End time')),
                ('location', models.CharField(blank=True, max_length=255)),
                ('details', wagtail.core.fields.RichTextField(blank=True)),
                ('website', models.URLField(blank=True)),
                ('repeat', ls.joyous.fields.RecurrenceField()),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.EventCategory', verbose_name='Category')),
                ('group_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='recurringeventpage_events', related_query_name='recurringeventpage_event', to='joyous.GroupPage')),
                ('image', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailimages.Image')),
            ],
            options={
                'verbose_name': 'Recurring Event Page',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='SimpleEventPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('time_from', models.TimeField(blank=True, null=True, verbose_name='Start time')),
                ('time_to', models.TimeField(blank=True, null=True, verbose_name='End time')),
                ('location', models.CharField(blank=True, max_length=255)),
                ('details', wagtail.core.fields.RichTextField(blank=True)),
                ('website', models.URLField(blank=True)),
                ('date', models.DateField(default=datetime.date.today, verbose_name='Date')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.EventCategory', verbose_name='Category')),
                ('group_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='simpleeventpage_events', related_query_name='simpleeventpage_event', to='joyous.GroupPage')),
                ('image', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailimages.Image')),
            ],
            options={
                'verbose_name': 'One-Off Event Page',
            },
            bases=('wagtailcore.page', models.Model),
        ),
        migrations.CreateModel(
            name='PostponementPage',
            fields=[
                ('cancellationpage_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='joyous.CancellationPage')),
                ('time_from', models.TimeField(blank=True, null=True, verbose_name='Start time')),
                ('time_to', models.TimeField(blank=True, null=True, verbose_name='End time')),
                ('location', models.CharField(blank=True, max_length=255)),
                ('details', wagtail.core.fields.RichTextField(blank=True)),
                ('website', models.URLField(blank=True)),
                ('postponement_title', models.CharField(help_text='The title for the postponed event', max_length=255, verbose_name='Title')),
                ('date', models.DateField(verbose_name='Date')),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.EventCategory', verbose_name='Category')),
                ('group_page', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='postponementpage_events', related_query_name='postponementpage_event', to='joyous.GroupPage')),
                ('image', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='wagtailimages.Image')),
            ],
            options={
                'verbose_name': 'Postponement',
            },
            bases=('joyous.cancellationpage', models.Model),
        ),
        migrations.AddField(
            model_name='extrainfopage',
            name='overrides',
            field=models.ForeignKey(help_text='The recurring event that we are updating.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.RecurringEventPage'),
        ),
        migrations.AddField(
            model_name='cancellationpage',
            name='overrides',
            field=models.ForeignKey(help_text='The recurring event that we are updating.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='joyous.RecurringEventPage'),
        ),
    ]



from django.contrib import admin
from ls.joyous.models import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, MultidayRecurringEventPage,
        CancellationPage, ExtraInfoPage, PostponementPage,
        RescheduleMultidayEventPage, GroupPage,
        CalendarPage, GeneralCalendarPage, SpecificCalendarPage)

admin.site.register(CalendarPage)
admin.site.register(SimpleEventPage, date_hierarchy = 'date')
admin.site.register(MultidayEventPage, date_hierarchy = 'date_from')
admin.site.register(RecurringEventPage)
admin.site.register(CancellationPage, date_hierarchy = 'except_date')
admin.site.register(ExtraInfoPage, date_hierarchy = 'except_date')
admin.site.register(PostponementPage, date_hierarchy = 'except_date')
admin.site.register(GroupPage)



# ------------------------------------------------------------------------------
# Test Widgets
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.test import TestCase
from wagtail.admin.widgets import AdminTimeInput, AdminDateInput
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import YEARLY, WEEKLY, MONTHLY
from ls.joyous.utils.recurrence import MO, TU, WE, TH, FR, SA, SU
from ls.joyous.widgets import RecurrenceWidget, Time12hrInput, ExceptionDateInput
from .testutils import datetimetz, freeze_timetz, getPage

# ------------------------------------------------------------------------------
class TestRecurrenceWidget(TestCase):
    def testDecompressNull(self):
        widget = RecurrenceWidget()
        self.assertEqual(widget.decompress(None),
                         [None, None, 1, [], None, None,      #5
                          101, 200, None, None, None,         #10
                          None, []])

    def testDecompressWeekdays(self):
        rr = Recurrence(dtstart=dt.date(2009, 1, 1),
                        freq=WEEKLY,
                        count=9,
                        byweekday=[MO,TU,WE,TH,FR])
        widget = RecurrenceWidget()
        self.assertEqual(widget.decompress(rr),
                         [dt.date(2009, 1, 1), WEEKLY, 1,
                          [0,1,2,3,4], 9, None,               #5
                          101, 200, None, None, None,         #10
                          None, []])

    def testDecompressEverydayInJanuary(self):
        rr = Recurrence(dtstart=dt.date(2014, 12, 1),
                        freq=YEARLY,
                        byweekday=[MO,TU,WE,TH,FR,SA,SU],
                        bymonth=[1])
        widget = RecurrenceWidget()
        self.assertEqual(widget.decompress(rr),
                         [dt.date(2014, 12, 1), YEARLY, 1,
                          [], None, None,                     #5
                          100, 200, None, None, None,         #10
                          None, [1]])

    def testNullValue(self):
        widget = RecurrenceWidget()
        self.assertEqual(widget.value_from_datadict({}, {}, 'repeat'), None)

    def testWeekdaysValue(self):
        widget = RecurrenceWidget()
        data = {'repeat_0':  '2009-01-01',
                'repeat_1':  '2',
                'repeat_2':  '1',
                'repeat_3':  ['0','1','2','3','4'],
                'repeat_5':  '2012-01-31',
                'repeat_6':  '101',
                'repeat_7':  '200',
                'repeat_8':  None,
                'repeat_9':  None,
                'repeat_10': None,
                'repeat_11': None,
                'repeat_12': ['1']}
        rr = Recurrence(dtstart=dt.date(2009, 1, 1),
                        freq=WEEKLY,
                        byweekday=[MO,TU,WE,TH,FR],
                        until=dt.date(2012,1,31))
        self.assertEqual(str(widget.value_from_datadict(data, {}, 'repeat')),
                         str(rr))

    def testEverydayInJanuaryValue(self):
        widget = RecurrenceWidget()
        data = {'repeat_0':  '2014-12-1',
                'repeat_1':  '0',
                'repeat_2':  '1',
                'repeat_5':  '',
                'repeat_6':  '100',
                'repeat_7':  '200',
                'repeat_8':  None,
                'repeat_9':  None,
                'repeat_10': None,
                'repeat_11': None,
                'repeat_12': ['1']}
        rr = Recurrence(dtstart=dt.date(2014, 12, 1),
                        freq=YEARLY,
                        byweekday=[MO,TU,WE,TH,FR,SA,SU],
                        bymonth=[1])
        self.assertEqual(str(widget.value_from_datadict(data, {}, 'repeat')),
                         str(rr))

    def testGetContext(self):
        rr = Recurrence(dtstart=dt.date(2014, 12, 1),
                        freq=MONTHLY,
                        byweekday=[MO(1),MO(3),MO(5)])
        attrs = {'id': "id_repeat"}
        widget = RecurrenceWidget(attrs)
        context = widget.get_context("repeat", rr, None)
        self.assertEqual(list(context.keys()), ["widget"])
        ctx = context['widget']
        self.assertEqual(ctx['attrs'], attrs)
        self.assertEqual(ctx['name'], "repeat")
        self.assertIsInstance(ctx['subwidgets'], list)
        self.assertEqual(len(ctx['subwidgets']), 13)
        self.assertEqual(ctx['template_name'],
                         "joyous/widgets/recurrence_widget.html")
        when = "The first Monday, third Monday and fifth Monday of the month"
        self.assertEqual(ctx['value'], when)
        self.assertEqual(ctx['value_s'], when)
        self.assertEqual(ctx['value_r'], "DTSTART:20141201\n"
                         "RRULE:FREQ=MONTHLY;WKST=SU;BYDAY=+1MO,+3MO,+5MO")

    def testMedia(self):
        widget = RecurrenceWidget()
        self.assertEqual(widget.media._css, {'all': ["/static/joyous/css/recurrence_admin.css"]})
        self.assertEqual(widget.media._js, ["/static/joyous/js/recurrence_admin.js"])

# ------------------------------------------------------------------------------
class TestTime12hrInput(TestCase):
    def setUp(self):
        self.newTime = AdminTimeInput().attrs.get('autocomplete', "new-time")

    def testNullValue(self):
        widget = Time12hrInput()
        self.assertEqual(widget.value_from_datadict({}, {}, 'time'), None)

    def testRenderNone(self):
        widget = Time12hrInput()
        out = widget.render('time', None, {'id': "time_id"})
        self.assertHTMLEqual(out, """
<input type="text" name="time" id="time_id" autocomplete="{0.newTime}">
<script>
$(function() {{
    initTime12hrChooser("time_id");
}});
</script>""".format(self))

    def testRenderValues(self):
        attrs = {'id': "time_id"}
        widget = Time12hrInput()
        out = widget.render('time', dt.time(10,15,54,89123), attrs)
        self.assertHTMLEqual(out, """
<input type="text" name="time" id="time_id" autocomplete="{0.newTime}" value="10:15am">
<script>
$(function() {{
    initTime12hrChooser("time_id");
}});
</script>""".format(self))
        out = widget.render('time', dt.time(12,51,34,89123), attrs)
        self.assertHTMLEqual(out, """
<input type="text" name="time" id="time_id" autocomplete="{0.newTime}" value="12:51pm">
<script>
$(function() {{
    initTime12hrChooser("time_id");
}});
</script>""".format(self))

    def testRenderFromString(self):
        attrs = {'id': "time_id"}
        widget = Time12hrInput()
        out = widget.render('time', "1pm", attrs)
        self.assertHTMLEqual(out, """
<input type="text" name="time" id="time_id" autocomplete="{0.newTime}" value="1pm">
<script>
$(function() {{
    initTime12hrChooser("time_id");
}});
</script>""".format(self))

    def testMedia(self):
        widget = Time12hrInput()
        self.assertEqual(widget.media._css, {})
        self.assertEqual(widget.media._js, ["/static/joyous/js/time12hr_admin.js"])

# ------------------------------------------------------------------------------
class TestExceptionDateInput(TestCase):
    def setUp(self):
        self.newDate = AdminDateInput().attrs.get('autocomplete', "new-date")

    def testNullValue(self):
        widget = ExceptionDateInput()
        self.assertEqual(widget.value_from_datadict({}, {}, 'xdate'), None)

    def testRenderNone(self):
        widget = ExceptionDateInput()
        out = widget.render('xdate', None, {'id': "id_xdate"})
        lines = [line for line in out.split("\n") if line]
        self.assertHTMLEqual(lines[0], """
<input type="text" name="xdate" id="id_xdate" autocomplete="{0.newDate}">""".format(self))
        self.assertIn('<script>initExceptionDateChooser("id_xdate", null, ', lines[1]);
        self.assertIn('"dayOfWeekStart": 0', lines[1])
        self.assertIn('"format": "Y-m-d"', lines[1])
        self.assertIn('</script>', lines[1])

    @freeze_timetz("2019-04-06 9:00")
    def testValidDates(self):
        widget = ExceptionDateInput()
        widget.overrides_repeat = Recurrence(dtstart=dt.date(2009, 1, 1),
                                             freq=MONTHLY, byweekday=MO(1))
        self.assertEqual(widget.valid_dates(),
                         ["20180903", "20181001", "20181105", "20181203", "20190107", "20190204",
                          "20190304", "20190401", "20190506", "20190603", "20190701", "20190805",
                          "20190902", "20191007", "20191104", "20191202", "20200106", "20200203",
                          "20200302", "20200406", "20200504", "20200601", "20200706", "20200803",
                          "20200907", "20201005" ])

    def testMedia(self):
        widget = ExceptionDateInput()
        self.assertEqual(widget.media._css, {'all': ["/static/joyous/css/recurrence_admin.css"]})
        self.assertEqual(widget.media._js, ["/static/joyous/js/recurrence_admin.js"])

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Recurrence
# ------------------------------------------------------------------------------
# Somewhat based upon RFC 5545 RRules, implemented using dateutil.rrule
# Does not support timezones ... and probably never will
# Does not support a frequency of by hour, by minute or by second
#
# See also:
#   https://github.com/django-recurrence/django-recurrence
#   https://github.com/dakrauth/django-swingtime

import sys
from operator import attrgetter
import calendar
import datetime as dt
from dateutil.rrule import rrule, rrulestr, rrulebase
from dateutil.rrule import DAILY, WEEKLY, MONTHLY, YEARLY
from dateutil.rrule import weekday as rrweekday
from django.utils.translation import gettext as _
from .telltime import dateFormatDMY
from .manythings import toOrdinal, toTheOrdinal, hrJoin
from .names import (WEEKDAY_NAMES, WEEKDAY_NAMES_PLURAL,
                    MONTH_NAMES, WRAPPED_MONTH_NAMES)

# ------------------------------------------------------------------------------
class Weekday(rrweekday):
    """
    Represents a day of the week, for every occurence of the week
    or for a specific week in the period.  e.g. The first Friday of the month.
    """
    def __repr__(self):
        s = ("MO", "TU", "WE", "TH", "FR", "SA", "SU")[self.weekday]
        if not self.n:
            return s
        else:
            return "{:+d}{}".format(self.n, s)

    def __str__(self):
        return self._getWhen(0)

    def _getWhen(self, offset, names=WEEKDAY_NAMES):
        weekday = names[self.weekday]
        if offset == 0:
            if not self.n:
                return weekday
            else:
                ordinal = toOrdinal(self.n)
                return _("{ordinal} {weekday}").format(**locals())

        localWeekday = names[(self.weekday + offset) % 7]
        if not self.n:
            return localWeekday
        else:
            ordinal = toOrdinal(self.n)
            if offset < 0:
                return _("{localWeekday} before the "
                         "{ordinal} {weekday}").format(**locals())
            else:
                return _("{localWeekday} after the "
                         "{ordinal} {weekday}").format(**locals())

    def _getPluralWhen(self, offset):
        return self._getWhen(offset, WEEKDAY_NAMES_PLURAL)

MO, TU, WE, TH, FR, SA, SU = EVERYWEEKDAY = map(Weekday, range(7))

# ------------------------------------------------------------------------------
class Recurrence(rrulebase):
    """
    Implementation of the recurrence rules somewhat based upon
    `RFC 5545 <https://tools.ietf.org/html/rfc5545>`_ RRules,
    implemented using dateutil.rrule.

    Does not support timezones ... and probably never will.
    Does not support a frequency of by-hour, by-minute or by-second.
    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        arg0 = args[0] if len(args) else None
        if isinstance(arg0, str):
            self.rule = rrulestr(arg0, **kwargs)
            if not isinstance(self.rule, rrule):
                raise ValueError("Only support simple RRules for now")
        elif isinstance(arg0, Recurrence):
            self.rule = arg0.rule
        elif isinstance(arg0, rrule):
            self.rule = arg0
        else:
            self.rule = rrule(*args, **kwargs)

    # expose all rrule properties
    #: How often the recurrence repeats. (0,1,2,3)
    freq = property(attrgetter("rule._freq"))

    #: The interval between each freq iteration.
    interval = property(attrgetter("rule._interval"))

    #: Limit on the number of occurrences.
    count = property(attrgetter("rule._count"))

    #: The week numbers to apply the recurrence to.
    byweekno = property(attrgetter("rule._byweekno"))

    #: The year days to apply the recurrence to.
    byyearday = property(attrgetter("rule._byyearday"))

    #: An offset from Easter Sunday.
    byeaster = property(attrgetter("rule._byeaster"))

    #: The nth occurrence of the rule inside the frequency period.
    bysetpos = property(attrgetter("rule._bysetpos"))

    @property
    def dtstart(self):
        """
        The recurrence start date.
        """
        return self.rule._dtstart.date()

    @property
    def frequency(self):
        """
        How often the recurrence repeats.
        ("YEARLY", "MONTHLY", "WEEKLY", "DAILY")
        """
        freqOptions = ("YEARLY", "MONTHLY", "WEEKLY", "DAILY")
        if self.rule._freq < len(freqOptions):
            return freqOptions[self.rule._freq]
        else:
            return "unsupported_frequency_{}".format(self.rule._freq)

    @property
    def until(self):
        """
        The last occurence in the rule is the greatest date that is
        less than or equal to the value specified in the until parameter.
        """
        if self.rule._until is not None:
            return self.rule._until.date()

    @property
    def wkst(self):
        """
        The week start day.  The default week start is got from
        calendar.firstweekday() which Joyous sets based on the Django
        FIRST_DAY_OF_WEEK setting.
        """
        return Weekday(self.rule._wkst)

    @property
    def byweekday(self):
        """
        The weekdays where the recurrence will be applied.  In RFC5545 this is
        called BYDAY, but is renamed by dateutil to avoid ambiguity.
        """
        retval = []
        if self.rule._byweekday:
            retval += [Weekday(day) for day in self.rule._byweekday]
        if self.rule._bynweekday:
            retval += [Weekday(day, n) for day, n in self.rule._bynweekday]
        return retval

    @property
    def bymonthday(self):
        """
        The month days where the recurrence will be applied.
        """
        retval = []
        if self.rule._bymonthday:
            retval += self.rule._bymonthday
        if self.rule._bynmonthday:
            retval += self.rule._bynmonthday
        return retval

    @property
    def bymonth(self):
        """
        The months where the recurrence will be applied.
        """
        if self.rule._bymonth:
            return list(self.rule._bymonth)
        else:
            return []

    def _iter(self):
        for occurence in self.rule._iter():
            yield occurence.date()

    # __len__() introduces a large performance penality.
    def getCount(self):
        """
        How many occurrences will be generated.
        """
        return self.rule.count()

    def __repr__(self):
        dtstart = ""
        if self.dtstart:
            dtstart = "DTSTART:{:%Y%m%d}\n".format(self.dtstart)
        rrule = "RRULE:{}".format(self._getRrule())
        retval = dtstart + rrule
        return retval

    def _getRrule(self, untilDt=None):
        # untilDt is the UTC datetime version of self.until
        if untilDt and untilDt.utcoffset() != dt.timedelta(0):
            raise TypeError("untilDt must be a UTC datetime")
        parts = ["FREQ={}".format(self.frequency)]
        if self.interval and self.interval != 1:
            parts.append("INTERVAL={}".format(self.interval))
        if self.wkst:
            parts.append("WKST={!r}".format(self.wkst))
        if self.count:
            parts.append("COUNT={}".format(self.count))
        if untilDt:
            parts.append("UNTIL={:%Y%m%dT%H%M%SZ}".format(untilDt))
        elif self.until:
            parts.append("UNTIL={:%Y%m%d}".format(self.until))
        for name, value in [('BYSETPOS',   self.bysetpos),
                            ('BYDAY',      self.byweekday),
                            ('BYMONTH',    self.bymonth),
                            ('BYMONTHDAY', self.bymonthday),
                            ('BYYEARDAY',  self.byyearday),
                            ('BYWEEKNO',   self.byweekno)]:
            if value:
                parts.append("{}={}".format(name,
                                            ",".join(repr(v) for v in value)))
        return ";".join(parts)

    def __str__(self):
        return self._getWhen(0)

    def _getWhen(self, offset, numDays=1):
        retval = ""
        if self.freq == DAILY:
            if self.interval > 1:
                retval = _("Every {n} days").format(n=self.interval)
            else:
                retval = _("Daily")
        elif self.freq == WEEKLY:
            retval = hrJoin([d._getPluralWhen(offset) for d in self.byweekday])
            if self.interval == 2:
                retval = _("Fortnightly on {days}").format(days=retval)
            elif self.interval > 2:
                retval = _("Every {n} weeks on {days}").format(n=self.interval,
                                                               days=retval)

        elif self.freq in (MONTHLY, YEARLY):
            if self.freq == MONTHLY:
                of = " "+_("of the month")
            else:
                months = hrJoin([MONTH_NAMES[m] for m in self.bymonth])
                of = " "+_("of {months}").format(months=months)
            days = []
            if self.byweekday:
                if (len(self.byweekday) == 7 and
                    all(not day.n for day in self.byweekday)):
                    retval = _("Everyday")
                    of = ""
                else:
                    retval = hrJoin([d._getWhen(offset) for d in self.byweekday])
                    if not self.byweekday[0].n:
                        retval = _("Every {when}").format(when=retval)
                        of = ""
                    else:
                        retval = _("The {when}").format(when=retval)

            elif len(self.bymonthday) > 1:
                days = [toTheOrdinal(d, False) for d in self.bymonthday]
                if offset == -2:
                    retval = _("Two days before {theOrdinal} day") \
                             .format(theOrdinal=hrJoin(days))
                elif offset == -1:
                    retval = _("The day before {theOrdinal} day")  \
                             .format(theOrdinal=hrJoin(days))
                elif offset == 1:
                    retval = _("The day after {theOrdinal} day")   \
                             .format(theOrdinal=hrJoin(days))
                elif offset == 2:
                    retval = _("Two days after {theOrdinal} day")  \
                             .format(theOrdinal=hrJoin(days))
                elif offset != 0:
                    retval = hrJoin(["{}{:+d}".format(day, offset)
                                     for day in days])

            elif len(self.bymonthday) == 1:
                d = self.bymonthday[0]
                if d == 1 and offset < 0:
                    d = offset
                    if self.freq != MONTHLY:
                        months = [WRAPPED_MONTH_NAMES[m-1] for m in self.bymonth]
                        of = " "+_("of {months}").format(months=hrJoin(months))
                elif d == -1 and offset > 0:
                    d = offset
                    if self.freq != MONTHLY:
                        months = [WRAPPED_MONTH_NAMES[m+1] for m in self.bymonth]
                        of = " "+_("of {months}").format(months=hrJoin(months))
                else:
                    d += offset
                retval = _("The {ordinal} day").format(ordinal=toOrdinal(d))
            retval += of
            if self.interval >= 2:
                if self.freq == MONTHLY:
                    retval = _("{when}, every {n} months")  \
                             .format(when=retval, n=self.interval)
                else:
                    retval = _("{when}, every {n} years")   \
                             .format(when=retval, n=self.interval)
        if numDays >= 2:
            retval += " "+_("for {n} days").format(n=numDays)
        if self.until:
            until = self.until + dt.timedelta(days=offset)
            # TODO make format configurable
            retval += " "+_("(until {when})").format(when=dateFormatDMY(until))
        return retval

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test ical Format
# ------------------------------------------------------------------------------
import sys
import datetime as dt
import pytz
from icalendar import vDatetime
from django.contrib.auth.models import User
from django.test import TestCase, RequestFactory
from wagtail.core.models import Site, Page
from ls.joyous.models.calendar import CalendarPage
from ls.joyous.models import (SimpleEventPage, MultidayEventPage,
        RecurringEventPage, MultidayRecurringEventPage,
        CancellationPage, ExtraInfoPage, PostponementPage)
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import DAILY, WEEKLY, MONTHLY, TU, SA
from ls.joyous.formats.ical import (SimpleVEvent, MultidayVEvent,
        RecurringVEvent, MultidayRecurringVEvent, ExtraInfoVEvent,
        PostponementVEvent, CancellationVEvent, VEventFactory, VEvent)
from freezegun import freeze_time

# ------------------------------------------------------------------------------
class TestBase(TestCase):
    def testMakePage(self):
        vev = VEvent()
        with self.assertRaises(NotImplementedError):
            vev.makePage()

# ------------------------------------------------------------------------------
class TestSimple(TestCase):
    def setUp(self):
        site = Site.objects.get(is_default_site=True)
        site.hostname = "joy.test"
        site.save()
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()

    def testFromPage(self):
        page = SimpleEventPage(owner = self.user,
                               slug  = "pet-show",
                               title = "Pet Show",
                               date  = dt.date(1987,6,5),
                               time_from = dt.time(11),
                               time_to   = dt.time(17,30),
                               tz = pytz.timezone("Australia/Sydney"))
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), SimpleVEvent)
        tz = pytz.timezone("Australia/Sydney")
        self.assertEqual(vev['DTSTART'].dt,
                         tz.localize(dt.datetime(1987,6,5,11,0)))

# ------------------------------------------------------------------------------
class TestMultiday(TestCase):
    def setUp(self):
        site = Site.objects.get(is_default_site=True)
        site.hostname = "joy.test"
        site.save()
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()

    def testFromPage(self):
        page = MultidayEventPage(owner = self.user,
                                 slug  = "niuekulele2018",
                                 title = "Niuekulele Ukulele Music Festival",
                                 date_from = dt.date(2018,3,16),
                                 date_to   = dt.date(2018,3,20),
                                 tz = pytz.timezone("Pacific/Niue"))
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), MultidayVEvent)
        tz = pytz.timezone("Pacific/Niue")
        self.assertEqual(vev['DTSTART'].dt,
                         tz.localize(dt.datetime(2018,3,16,0,0)))
        self.assertEqual(vev['DTEND'].dt,
                         tz.localize(dt.datetime(2018,3,20,23,59,59,999999)))

# ------------------------------------------------------------------------------
class TestRecurring(TestCase):
    def setUp(self):
        site = Site.objects.get(is_default_site=True)
        site.hostname = "joy.test"
        site.save()
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()

    @freeze_time("2017-08-15")
    def testFromPage(self):
        page = RecurringEventPage(owner = self.user,
                                  slug  = "code-for-boston",
                                  title = "Code for Boston",
                                  repeat    = Recurrence(dtstart=dt.date(2017,1,1),
                                                         freq=WEEKLY,
                                                         byweekday=[TU],
                                                         until=dt.date(2017,12,26)),
                                  time_from = dt.time(19),
                                  time_to   = dt.time(21,30),
                                  tz = pytz.timezone("US/Eastern"),
                                  location  = "4th Floor, 1 Broadway, Cambridge, MA")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), RecurringVEvent)
        vev.set('UID', "this-is-not-a-unique-identifier")
        codeForBoston = b"\r\n".join([
                b"BEGIN:VEVENT",
                b"SUMMARY:Code for Boston",
                b"DTSTART;TZID=US/Eastern:20170103T190000",
                b"DTEND;TZID=US/Eastern:20170103T213000",
                b"DTSTAMP:20170815T000000Z",
                b"UID:this-is-not-a-unique-identifier",
                b"SEQUENCE:1",
                b"RRULE:FREQ=WEEKLY;UNTIL=20171227T045959Z;BYDAY=TU;WKST=SU",
                b"CREATED:20170815T000000Z",
                b"DESCRIPTION:",
                b"LAST-MODIFIED:20170815T000000Z",
                b"LOCATION:4th Floor\\, 1 Broadway\\, Cambridge\\, MA",
                b"URL:http://joy.test/events/code-for-boston/",
                b"END:VEVENT",
                b""])
        self.assertEqual(vev.to_ical(), codeForBoston)

    @freeze_time("2018-05-10")
    def testExdate(self):
        page = RecurringEventPage(owner = self.user,
                                  slug  = "sleep",
                                  title = "Sleep In",
                                  repeat    = Recurrence(dtstart=dt.date(2018,5,1),
                                                         freq=MONTHLY,
                                                         byweekday=[SA(+2)]),
                                  time_from = dt.time(7),
                                  time_to   = dt.time(10,30),
                                  tz = pytz.timezone("Pacific/Auckland"),
                                  details = "<p>zzzZZZZZZZZZ</p>",
                                  location  = "Bed")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        except1 = CancellationPage(owner = self.user,
                                   slug  = "2018-06-09-cancellation",
                                   title = "Cancellation for Saturday 9th of June",
                                   overrides = page,
                                   except_date = dt.date(2018,6,9))
        page.add_child(instance=except1)
        except1.save_revision().publish()
        except2 = CancellationPage(owner = self.user,
                                   slug  = "2018-07-14-cancellation",
                                   title = "Cancellation for Saturday 14th of July",
                                   overrides = page,
                                   except_date = dt.date(2018,7,14))
        page.add_child(instance=except2)
        except2.save_revision().publish()
        vev = VEventFactory().makeFromPage(except1)
        self.assertIs(type(vev), RecurringVEvent)
        vev.set('UID', "this-is-not-a-unique-identifier")
        tz = pytz.timezone("Pacific/Auckland")
        exDates = [exDate.dt for exDate in vev['EXDATE'].dts]
        self.assertEqual(exDates, [tz.localize(dt.datetime(2018,6,9,7)),
                                   tz.localize(dt.datetime(2018,7,14,7))])
        sleepIn = b"\r\n".join([b"BEGIN:VEVENT",
                                b"SUMMARY:Sleep In",
                                b"DTSTART;TZID=Pacific/Auckland:20180512T070000",
                                b"DTEND;TZID=Pacific/Auckland:20180512T103000",
                                b"DTSTAMP:20180510T000000Z",
                                b"UID:this-is-not-a-unique-identifier",
                                b"SEQUENCE:1",
                                b"RRULE:FREQ=MONTHLY;BYDAY=+2SA;WKST=SU",
                                b"EXDATE;TZID=Pacific/Auckland:20180609T070000,20180714T070000",
                                b"CREATED:20180510T000000Z",
                                b"DESCRIPTION:<p>zzzZZZZZZZZZ</p>",
                                b"LAST-MODIFIED:20180510T000000Z",
                                b"LOCATION:Bed",
                                b"URL:http://joy.test/events/sleep/",
                                b"END:VEVENT",
                                b""])
        self.assertEqual(vev.to_ical(), sleepIn)

    @freeze_time("2017-05-15")
    def testExtraInfo(self):
        page = RecurringEventPage(owner = self.user,
                                  slug  = "code-for-boston",
                                  title = "Code for Boston",
                                  repeat    = Recurrence(dtstart=dt.date(2017,1,1),
                                                         freq=WEEKLY,
                                                         byweekday=[TU],
                                                         until=dt.date(2017,12,26)),
                                  time_from = dt.time(19),
                                  time_to   = dt.time(21,30),
                                  tz = pytz.timezone("US/Eastern"),
                                  location  = "4th Floor, 1 Broadway, Cambridge, MA")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        info = ExtraInfoPage(owner = self.user,
                             slug  = "2017-06-06-extra-info",
                             title = "Extra Information for Tuesday 6th of June",
                             overrides = page,
                             except_date = dt.date(2017,6,6),
                             extra_title = "Handling Time Zones with Python")
        page.add_child(instance=info)
        info.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), RecurringVEvent)
        vev.set('UID', "this-is-not-a-unique-identifier")
        codeForBoston = b"\r\n".join([
                b"BEGIN:VEVENT",
                b"SUMMARY:Code for Boston",
                b"DTSTART;TZID=US/Eastern:20170103T190000",
                b"DTEND;TZID=US/Eastern:20170103T213000",
                b"DTSTAMP:20170515T000000Z",
                b"UID:this-is-not-a-unique-identifier",
                b"SEQUENCE:1",
                b"RRULE:FREQ=WEEKLY;UNTIL=20171227T045959Z;BYDAY=TU;WKST=SU",
                b"CREATED:20170515T000000Z",
                b"DESCRIPTION:",
                b"LAST-MODIFIED:20170515T000000Z",
                b"LOCATION:4th Floor\\, 1 Broadway\\, Cambridge\\, MA",
                b"URL:http://joy.test/events/code-for-boston/",
                b"END:VEVENT",
                b""])
        self.assertEqual(vev.to_ical(), codeForBoston)
        self.assertEqual(len(vev.vchildren), 1)
        vchild = vev.vchildren[0]
        vchild.set('UID', "this-is-not-a-unique-identifier")
        self.assertIs(type(vchild), ExtraInfoVEvent)
        talk = b"\r\n".join([
                b"BEGIN:VEVENT",
                b"SUMMARY:Handling Time Zones with Python",
                b"DTSTART;TZID=US/Eastern:20170606T190000",
                b"DTEND;TZID=US/Eastern:20170606T213000",
                b"DTSTAMP:20170515T000000Z",
                b"UID:this-is-not-a-unique-identifier",
                b"RECURRENCE-ID;TZID=US/Eastern:20170606T190000",
                b"SEQUENCE:1",
                b"CREATED:20170515T000000Z",
                b"DESCRIPTION:",
                b"LAST-MODIFIED:20170515T000000Z",
                b"LOCATION:4th Floor\\, 1 Broadway\\, Cambridge\\, MA",
                b"URL:http://joy.test/events/code-for-boston/2017-06-06-extra-info/" ,
                b"END:VEVENT",
                b""])
        self.assertEqual(vchild.to_ical(), talk)

    @freeze_time("2017-05-15")
    def testPostponement(self):
        page = RecurringEventPage(owner = self.user,
                                  slug  = "code-for-boston",
                                  title = "Code for Boston",
                                  repeat    = Recurrence(dtstart=dt.date(2017,1,1),
                                                         freq=WEEKLY,
                                                         byweekday=[TU],
                                                         until=dt.date(2017,12,26)),
                                  time_from = dt.time(19),
                                  time_to   = dt.time(21,30),
                                  tz = pytz.timezone("US/Eastern"),
                                  location  = "4th Floor, 1 Broadway, Cambridge, MA")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        post = PostponementPage(owner = self.user,
                                slug  = "2017-06-13-postponement",
                                title = "Postponement for Tuesday 13th of June",
                                postponement_title = "Code for Boston (on Weds)",
                                overrides = page,
                                except_date = dt.date(2017,6,13),
                                date = dt.date(2017,6,14)
                                )
        page.add_child(instance=post)
        post.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), RecurringVEvent)
        self.assertEqual(len(vev.vchildren), 1)
        vchild = vev.vchildren[0]
        vchild.set('UID', "this-is-not-a-unique-identifier")
        self.assertIs(type(vchild), PostponementVEvent)
        talk = b"\r\n".join([
                b"BEGIN:VEVENT",
                b"SUMMARY:Code for Boston (on Weds)",
                b"DTSTART;TZID=US/Eastern:20170614T000000",
                b"DTEND;TZID=US/Eastern:20170614T235959",
                b"DTSTAMP:20170515T000000Z",
                b"UID:this-is-not-a-unique-identifier",
                b"RECURRENCE-ID;TZID=US/Eastern:20170613T190000",
                b"SEQUENCE:1",
                b"CREATED:20170515T000000Z",
                b"DESCRIPTION:",
                b"LAST-MODIFIED:20170515T000000Z",
                b"LOCATION:",
                b"URL:http://joy.test/events/code-for-boston/2017-06-13-postponement/",
                b"END:VEVENT",
                b""])
        self.assertEqual(vchild.to_ical(), talk)

    @freeze_time("2017-05-15")
    def testCancellation(self):
        # NOTE:
        # Cancellations are represented by EXDATE in VEVENT not as their own
        # VEVENT instance.  Yes, this means the cancellation_title and
        # cancellation_details are lost.
        page = RecurringEventPage(owner = self.user,
                                  slug  = "sleep",
                                  title = "Sleep In",
                                  repeat    = Recurrence(dtstart=dt.date(2018,5,1),
                                                         freq=MONTHLY,
                                                         byweekday=[SA(+2)]),
                                  time_from = dt.time(7),
                                  time_to   = dt.time(10,30),
                                  tz = pytz.timezone("Pacific/Auckland"),
                                  details = "<p>zzzZZZZZZZZZ</p>",
                                  location  = "Bed")
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        except1 = CancellationPage(owner = self.user,
                                   slug  = "2018-06-09-cancellation",
                                   title = "Cancellation for Saturday 9th of June",
                                   cancellation_title = "Get up early",
                                   overrides = page,
                                   except_date = dt.date(2018,6,9))
        page.add_child(instance=except1)
        except1.save_revision().publish()
        vexcept = CancellationVEvent.fromPage(except1)
        self.assertEqual(vexcept['SUMMARY'], "Get up early")
        self.assertEqual(vexcept['DESCRIPTION'], "")
        self.assertEqual(vexcept.to_ical(), b"\r\n")

# ------------------------------------------------------------------------------
class TestMultidayRecurring(TestCase):
    def setUp(self):
        site = Site.objects.get(is_default_site=True)
        site.hostname = "joy.test"
        site.save()
        self.home = Page.objects.get(slug='home')
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.calendar = CalendarPage(owner = self.user,
                                     slug  = "events",
                                     title = "Events")
        self.home.add_child(instance=self.calendar)
        self.calendar.save_revision().publish()

    @freeze_time("2017-08-15")
    def testFromPage(self):
        page = MultidayRecurringEventPage(owner = self.user,
                                  slug   = "bought-from-rubber-man",
                                  title  = "Bought from a Rubber Man",
                                  repeat = Recurrence(dtstart=dt.date(2019,4,2),
                                                      freq=WEEKLY,
                                                      byweekday=[TU]),
                                  num_days  = 3,
                                  time_from = dt.time(16),
                                  time_to   = dt.time(18),
                                  tz = pytz.timezone("Pacific/Auckland"))
        self.calendar.add_child(instance=page)
        page.save_revision().publish()
        vev = VEventFactory().makeFromPage(page)
        self.assertIs(type(vev), MultidayRecurringVEvent)
        tz = pytz.timezone("Pacific/Auckland")
        self.assertEqual(vev['DTSTART'].dt,
                         tz.localize(dt.datetime(2019,4,2,16,0)))
        self.assertEqual(vev['DTEND'].dt,
                         tz.localize(dt.datetime(2019,4,4,18,0)))
        self.assertEqual(vev['RRULE']['FREQ'],  ["WEEKLY"])
        self.assertEqual(vev['RRULE']['BYDAY'], ["TU"])

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# Generated by Django 2.0.7 on 2018-07-15 06:02

from django.db import migrations
from uuid import uuid4

def populate(model):
    def gen_uid(apps, schema_editor):
        Model = apps.get_model("joyous", model)
        for row in Model.objects.all():
            row.uid = uuid4()
            row.save(update_fields=["uid"])
    return gen_uid


class Migration(migrations.Migration):

    dependencies = [
        ('joyous', '0007_add_uid'),
    ]

    operations = [
        migrations.RunPython(populate("MultidayEventPage"),  migrations.RunPython.noop),
        migrations.RunPython(populate("PostponementPage"),   migrations.RunPython.noop),
        migrations.RunPython(populate("RecurringEventPage"), migrations.RunPython.noop),
        migrations.RunPython(populate("SimpleEventPage"),    migrations.RunPython.noop),
    ]



# ------------------------------------------------------------------------------
# Test Joyous Template Tags
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.contrib.auth.models import User
from django.test import TestCase, RequestFactory
from django.utils import timezone
from django.template import Context, Template, TemplateSyntaxError
from django.test import TestCase, RequestFactory
from wagtail.core.models import Site, Page
from bs4 import BeautifulSoup
from ls.joyous.utils.recurrence import Recurrence
from ls.joyous.utils.recurrence import DAILY, WEEKLY, YEARLY, MO, TU, WE, TH, FR, SA
from ls.joyous.models import (CalendarPage, SimpleEventPage, RecurringEventPage,
                              CancellationPage, PostponementPage, GroupPage)
from .testutils import datetimetz, freeze_timetz
from .testutils import getPage

# ------------------------------------------------------------------------------
class TestMultiSite(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        self.requestFactory = RequestFactory()
        self._setUpMainSite()
        self._setUpNovaSubsite()

    def _setUpMainSite(self):
        Site.objects.filter(is_default_site=True).update(hostname="joy.test")
        home = getPage("/home/")
        events = CalendarPage(owner = self.user,
                              slug  = "events",
                              title = "Events")
        home.add_child(instance=events)
        events.save_revision().publish()
        chess = GroupPage(slug="chess-club", title="Chess Club")
        home.add_child(instance=chess)
        chess.save_revision().publish()
        event = RecurringEventPage(owner = self.user,
                                   slug  = "lunchtime-matches",
                                   title = "Lunchtime Chess Matches",
                                   repeat = Recurrence(dtstart=dt.date(1984,8,5),
                                                       freq=WEEKLY,
                                                       byweekday=[MO,WE,FR]),
                                   time_from = dt.time(12),
                                   time_to   = dt.time(13))
        chess.add_child(instance=event)
        event.save_revision().publish()
        cancellation = CancellationPage(owner = self.user,
                                        slug  = "1984-10-01-cancellation",
                                        title = "Cancellation for Monday 1st of October",
                                        overrides = event,
                                        except_date = dt.date(1984, 10, 1),
                                        cancellation_title   = "No Chess Club Today")
        event.add_child(instance=cancellation)
        cancellation.save_revision().publish()
        postponement = PostponementPage(owner = self.user,
                                        slug  = "1984-10-03-postponement",
                                        title = "Postponement for Wednesday 3rd of October",
                                        overrides = event,
                                        except_date = dt.date(1984, 10, 3),
                                        cancellation_title   = "",
                                        postponement_title   = "Early Morning Matches",
                                        date      = dt.date(1984,10,4),
                                        time_from = dt.time(7,30),
                                        time_to   = dt.time(8,30))
        event.add_child(instance=postponement)
        postponement.save_revision().publish()
        event = RecurringEventPage(owner = self.user,
                                   slug  = "flea-market",
                                   title = "Flea Market",
                                   repeat = Recurrence(dtstart=dt.date(1984,7,1),
                                                       freq=YEARLY,
                                                       byweekday=[SA(1), SA(3)],
                                                       bymonth=range(2,12)),
                                   time_from = dt.time(8),
                                   time_to   = dt.time(13))
        events.add_child(instance=event)
        event.save_revision().publish()
        event = RecurringEventPage(owner = self.user,
                                   slug  = "drama-practice",
                                   title = "Drama Group",
                                   repeat = Recurrence(dtstart=dt.date(1984,8,14),
                                                       freq=WEEKLY,
                                                       byweekday=[TH],
                                                       interval=2),
                                   time_from = dt.time(17))
        events.add_child(instance=event)
        event.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "public-lecture3",
                                title = "The Human Environment",
                                date  = dt.date(1984,9,14),
                                time_from = dt.time(19),
                                location = "Lecture Hall C")
        events.add_child(instance=event)
        event.save_revision().publish()

    def _setUpNovaSubsite(self):
        main = getPage("/home/")
        home = Page(slug="nova", title="Nova Homepage")
        main.add_child(instance=home)
        home.save_revision().publish()
        activities = Page(slug="activities", title="Nova Activities")
        home.add_child(instance=activities)
        activities.save_revision().publish()
        Site.objects.create(hostname='nova.joy.test',
                            root_page_id=home.id,
                            is_default_site=False)
        events = CalendarPage(owner = self.user,
                              slug  = "nova-events",
                              title = "Nova Events")
        home.add_child(instance=events)
        events.save_revision().publish()
        committee = RecurringEventPage(owner = self.user,
                                       slug  = "executive-meeting",
                                       title = "Executive Committee Meeting",
                                       repeat = Recurrence(dtstart=dt.date(1984,8,5),
                                                           freq=WEEKLY,
                                                           byweekday=[TH]),
                                       time_from = dt.time(13),
                                       time_to   = dt.time(15))
        events.add_child(instance=committee)
        committee.save_revision().publish()
        event = SimpleEventPage(owner = self.user,
                                slug  = "rubbish-blitz",
                                title = "Rubbish Blitz",
                                date  = dt.date(1984, 9, 13),
                                time_from = dt.time(12,30),
                                time_to   = dt.time(17))
        events.add_child(instance=event)
        event.save_revision().publish()
        cancellation = CancellationPage(owner = self.user,
                                        slug  = "1984-09-13-cancellation",
                                        title = "Cancellation for Thursday 13th of September",
                                        overrides = committee,
                                        except_date = dt.date(1984, 9, 13),
                                        cancellation_title   = "Meeting Cancelled",
                                        cancellation_details = "The committee will be at "
                                                               "the working bee")
        committee.add_child(instance=cancellation)
        cancellation.save_revision().publish()

    def _getContext(self, hostname=None, slug="", **kwargs):
        if hostname is None:
            hostname = "joy.test"
        site = Site.objects.get(hostname=hostname)
        vars = {'request': self._getRequest(site, slug),
                'page':    getPage("/home/")}
        vars.update(kwargs)
        context = Context(vars)
        return context

    def _getRequest(self, site, slug=""):
        path = "test/{}/{}".format(site.root_page.slug, slug)
        request = self.requestFactory.get(path)
        request.user = self.user
        request.site = site
        request.session = {}
        return request

    @freeze_timetz("1984-09-11 10:00")
    def testEventsThisWeek(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% events_this_week %}"
        ).render(self._getContext())
        self.assertHTMLEqual(out, """
<div class="events-this-week">
  <h3>This Week</h3>
  <div class="events">
      <div class="day in-past">
        <div class="event-day">
            <h4>Monday</h4> 10th Sep
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
        </div>
      </div>
      <div class="day today">
        <div class="event-day" title="Today">
            <h4>Tuesday</h4> 11th Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day">
            <h4>Wednesday</h4> 12th Sep
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Thursday</h4> 13th Sep
        </div>
        <div class="days-events">
            <a href="/nova/nova-events/rubbish-blitz/" class="event">
              12:30pm Rubbish Blitz
            </a>
            <a href="/nova/nova-events/executive-meeting/1984-09-13-cancellation/" class="event">
              1pm Meeting Cancelled
            </a>
            <a href="/events/drama-practice/" class="event">
              5pm Drama Group
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Friday</h4> 14th Sep
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
            <a href="/events/public-lecture3/" class="event">
              7pm The Human Environment
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Saturday</h4> 15th Sep
        </div>
        <div class="days-events">
            <a href="/events/flea-market/" class="event">
              8am Flea Market
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Sunday</h4> 16th Sep
        </div>
        <div class="days-events">
        </div>
      </div>
  </div>
  <div class="calendar-link">
    <a href="/events/">Events</a>
  </div>
</div>
""")

    @freeze_timetz("1984-09-20 09:00")
    def testSubsiteEventsThisWeek(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% events_this_week %}"
        ).render(self._getContext(hostname="nova.joy.test",
                                  page=getPage("/home/nova/activities/")))
        self.assertHTMLEqual(out, """
<div class="events-this-week">
  <h3>This Week</h3>
  <div class="events">
      <div class="day in-past">
        <div class="event-day">
            <h4>Monday</h4> 17th Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day in-past">
        <div class="event-day">
            <h4>Tuesday</h4> 18th Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day in-past">
        <div class="event-day">
            <h4>Wednesday</h4> 19th Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day today">
        <div class="event-day" title="Today">
            <h4>Thursday</h4> 20th Sep
        </div>
        <div class="days-events">
            <a href="/nova-events/executive-meeting/" class="event">
              1pm Executive Committee Meeting
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Friday</h4> 21st Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Saturday</h4> 22nd Sep
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Sunday</h4> 23rd Sep
        </div>
        <div class="days-events">
        </div>
      </div>
  </div>
  <div class="calendar-link">
    <a href="/nova-events/">Nova Events</a>
  </div>
</div>
""")

    @freeze_timetz("1984-10-24 10:00")
    def testMinicalendar(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% minicalendar %}"
        ).render(self._getContext())
        soup = BeautifulSoup(out, "html5lib")
        select = soup.select
        self.assertEqual(len(select('thead a')), 2)
        self.assertEqual(len(select("table.minicalendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "October")
        year = select("tr.heading th.month .year-number")[0]
        self.assertEqual(year.string.strip(), "1984")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 31)
        self.assertEqual(len(select("tbody td.noday")), 4)
        self.assertEqual(len(select('tbody td.day a[title="Lunchtime Chess Matches"]')), 12)
        links = select('tbody td.day a')
        self.assertEqual(len(links), 19)
        self.assertEqual(links[0].get_text(), "1")
        self.assertEqual(links[0]['href'], "/events/1984/10/01/")
        self.assertEqual(links[0]['title'], "No Chess Club Today")
        self.assertEqual(links[1].get_text(), "4")
        self.assertEqual(links[1]['href'], "/events/1984/10/04/")
        self.assertEqual(links[1]['title'], "Early Morning Matches, Executive Committee Meeting")
        self.assertEqual(links[3].get_text(), "6")
        self.assertEqual(links[3]['href'], "/events/1984/10/06/")
        self.assertEqual(links[3]['title'], "Flea Market")
        self.assertEqual(links[15].get_text(), "25")
        self.assertEqual(links[15]['href'], "/events/1984/10/25/")
        self.assertEqual(links[15]['title'], "Executive Committee Meeting, Drama Group")

    @freeze_timetz("1984-09-05 10:00")
    def testAllUpcomingEvents(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% all_upcoming_events %}"
        ).render(self._getContext())
        soup = BeautifulSoup(out, "html5lib")
        items = soup(class_="event-item")
        self.assertEqual(len(items), 7)
        chess, novaexec, blitz, drama, lecture, market, chess2 = items
        self.assertEqual(chess.a['href'], "/chess-club/lunchtime-matches/")
        self.assertEqual(novaexec.find(class_="event-next-on").get_text(strip=True),
                         "Next on Thursday 6th of September at 1pm")
        self.assertEqual(blitz.a['href'], "/nova/nova-events/rubbish-blitz/")
        self.assertEqual(blitz.a.get_text(strip=True), "Rubbish Blitz")
        self.assertEqual(blitz.find(class_="event-when").get_text(strip=True),
                         "Thursday 13th of September at 12:30pm to 5pm")
        self.assertEqual(drama.find(class_="event-when").get_text(strip=True),
                         "Fortnightly on Thursdays at 5pm")
        self.assertEqual(drama.find(class_="event-next-on").get_text(strip=True),
                         "Next on Thursday 13th of September at 5pm")
        self.assertEqual(lecture.find(class_="event-when").get_text(strip=True),
                         "Friday 14th of September at 7pm")
        self.assertEqual(lecture.find(class_="event-location").get_text(strip=True),
                         "Lecture Hall C")
        self.assertEqual(market.a['href'], "/events/flea-market/")
        self.assertEqual(market.a.get_text(strip=True), "Flea Market")
        self.assertEqual(chess2.a['href'], "/chess-club/lunchtime-matches/1984-10-03-postponement/")
        self.assertEqual(chess2.find(class_="event-postponed-from").get_text(strip=True),
                         "Postponed from Wednesday 3rd of October at 12pm to 1pm")

    @freeze_timetz("1984-09-01 15:00")
    def testSubsiteUpcomingEvents(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% subsite_upcoming_events %}"
        ).render(self._getContext(hostname="nova.joy.test",
                                  page=getPage("/home/nova/activities/")))
        soup = BeautifulSoup(out, "html5lib")
        items = soup(class_="event-item")
        self.assertEqual(len(items), 2)
        novaexec, blitz = items
        self.assertEqual(novaexec.a.get_text(strip=True),
                         "Executive Committee Meeting")
        self.assertEqual(blitz.a.get_text(strip=True), "Rubbish Blitz")

    @freeze_timetz("1984-09-15 10:00")
    def testGroupUpcomingEvents(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% group_upcoming_events %}"
        ).render(self._getContext(page=getPage("/home/chess-club/")))
        self.assertHTMLEqual(out, """
<h3>Events</h3>
<div class="upcoming-events">
  <div class="event-item clearfix">
    <div class="event-item-title">
      <a href="/chess-club/lunchtime-matches/">Lunchtime Chess Matches</a>
    </div>
    <div class="event-when">
      Mondays, Wednesdays and Fridays at 12pm to 1pm
    </div>
    <div class="event-next-on">
      Next on Monday 17th of September at 12pm
    </div>
  </div>
  <div class="event-item clearfix">
    <div class="event-item-title">
      <a href="/chess-club/lunchtime-matches/1984-10-03-postponement/">Early Morning Matches</a>
    </div>
    <div class="event-when">
      Thursday 4th of October at 7:30am to 8:30am
    </div>
    <div class="event-postponed-from">
      Postponed from Wednesday 3rd of October at 12pm to 1pm
    </div>
  </div>
</div>""")

    @freeze_timetz("1984-08-04 10:00")
    def testFutureExceptionsImplicitPage(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% future_exceptions %}"
        ).render(self._getContext(hostname="nova.joy.test",
                                  page=getPage("/home/nova/nova-events/executive-meeting/")))
        soup = BeautifulSoup(out, "html5lib")
        items = soup(class_="item")
        self.assertEqual(len(items), 1)
        cancellation = items[0]
        self.assertEqual(cancellation['href'],
                         "/nova-events/executive-meeting/1984-09-13-cancellation/")
        self.assertEqual(cancellation.find(class_="item-text").get_text(strip=True),
                         "Meeting Cancelled")

    @freeze_timetz("1984-09-15 13:00")
    def testFutureExceptionsExplicitPage(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% future_exceptions kasparov %}"
        ).render(self._getContext(kasparov=getPage("/home/chess-club/lunchtime-matches/")))
        soup = BeautifulSoup(out, "html5lib")
        items = soup(class_="item")
        self.assertEqual(len(items), 2)
        cancellation, postponement = items
        self.assertEqual(cancellation['href'],
                         "/chess-club/lunchtime-matches/1984-10-01-cancellation/")
        self.assertEqual(cancellation.find(class_="item-text").get_text(strip=True),
                         "No Chess Club Today")
        self.assertEqual(postponement['href'],
                         "/chess-club/lunchtime-matches/1984-10-03-postponement/")
        self.assertEqual(postponement.find(class_="item-text").get_text(strip=True),
                         "Early Morning Matches on Thursday 4th of October")

# ------------------------------------------------------------------------------
class TestNoCalendar(TestCase):
    def setUp(self):
        self.user = User.objects.create_user('i', 'i@joy.test', 's3cr3t')
        home = getPage("/home/")
        chess = GroupPage(slug="chess-club", title="Chess Club")
        home.add_child(instance=chess)
        chess.save_revision().publish()
        event = RecurringEventPage(owner = self.user,
                                   slug  = "lunchtime-matches",
                                   title = "Lunchtime Chess Matches",
                                   repeat = Recurrence(dtstart=dt.date(1984,8,5),
                                                       freq=WEEKLY,
                                                       byweekday=[MO,WE,FR]),
                                   time_from = dt.time(12),
                                   time_to   = dt.time(13))
        chess.add_child(instance=event)
        event.save_revision().publish()
        self.request = RequestFactory().get("/test")
        self.request.user = self.user
        self.request.session = {}
        self.request.site = Site.objects.get(is_default_site=True)

    @freeze_timetz("1987-08-16 12:45")
    def testEventsThisWeek(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% events_this_week %}"
        ).render(Context({'request': self.request}))
        self.assertHTMLEqual(out, """
<div class="events-this-week">
  <h3>This Week</h3>
  <div class="events">
      <div class="day today">
        <div class="event-day" title="Today">
            <h4>Sunday</h4> 16th Aug
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day">
            <h4>Monday</h4> 17th Aug
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day">
            <h4>Tuesday</h4> 18th Aug
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day">
            <h4>Wednesday</h4> 19th Aug
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Thursday</h4> 20th Aug
        </div>
        <div class="days-events">
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Friday</h4> 21st Aug
        </div>
        <div class="days-events">
            <a href="/chess-club/lunchtime-matches/" class="event">
              12pm Lunchtime Chess Matches
            </a>
        </div>
      </div>
      <div class="day">
        <div class="event-day" >
            <h4>Saturday</h4> 22nd Aug
        </div>
        <div class="days-events">
        </div>
      </div>
  </div>
</div>
""")

    @freeze_timetz("1984-10-24 10:00")
    def testMinicalendar(self):
        out = Template(
            "{% load joyous_tags %}"
            "{% minicalendar %}"
        ).render(Context({'request': self.request}))
        soup = BeautifulSoup(out, "html5lib")
        select = soup.select
        self.assertEqual(len(select('thead a')), 0)
        self.assertEqual(len(select("table.minicalendar thead tr th.sun")), 1)
        month = select("tr.heading th.month .month-name")[0]
        self.assertEqual(month.string.strip(), "October")
        year = select("tr.heading th.month .year-number")[0]
        self.assertEqual(year.string.strip(), "1984")
        self.assertEqual(len(select("tbody tr")), 5)
        self.assertEqual(len(select("tbody td")), 35)
        self.assertEqual(len(select("tbody td.day")), 31)
        self.assertEqual(len(select("tbody td.noday")), 4)
        self.assertEqual(len(select('tbody td.day span.event')), 14)

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Test Recurrence
# ------------------------------------------------------------------------------
import sys
import datetime as dt
from django.test import TestCase
from ls.joyous.utils.recurrence import Recurrence, Weekday
from ls.joyous.utils.recurrence import MO, TU, WE, TH, FR, SA, SU
from ls.joyous.utils.recurrence import YEARLY, MONTHLY, WEEKLY, DAILY

# ------------------------------------------------------------------------------
class TestWeekday(TestCase):
    def testStr(self):
        self.assertEqual(str(Weekday(0)), "Monday")
        self.assertEqual(str(Weekday(4,1)), "first Friday")
        self.assertEqual(str(Weekday(4,-1)), "last Friday")
        self.assertEqual(str(SA), "Saturday")
        self.assertEqual(str(FR(3)), "third Friday")

    def testGetWhen(self):
        self.assertEqual(Weekday(0)._getWhen(0), "Monday")
        self.assertEqual(FR(1)._getWhen(0), "first Friday")
        self.assertEqual(SU._getWhen(1), "Monday")
        self.assertEqual(WE._getWhen(-2), "Monday")
        self.assertEqual(FR(1)._getWhen(-1), "Thursday before the first Friday")
        self.assertEqual(SU(1)._getWhen(2), "Tuesday after the first Sunday")

    def testRepr(self):
        self.assertEqual(repr(Weekday(0)), "MO")
        self.assertEqual(repr(Weekday(4,2)), "+2FR")
        self.assertEqual(repr(SA), "SA")
        self.assertEqual(repr(FR(3)), "+3FR")
        self.assertEqual(repr(WE(-2)), "-2WE")

# ------------------------------------------------------------------------------
class TestRecurrence(TestCase):
    def testRepr(self):
        rr = Recurrence(dtstart=dt.date(2009, 1, 1),
                        freq=WEEKLY,
                        count=9,
                        byweekday=[MO,TU,WE,TH,FR])
        self.assertEqual(repr(rr),
                        "DTSTART:20090101\n" \
                        "RRULE:FREQ=WEEKLY;WKST=SU;COUNT=9;BYDAY=MO,TU,WE,TH,FR")
        self.assertEqual(rr.count, rr.getCount())
        rr = Recurrence(dtstart=dt.date(2011, 1, 1),
                        freq=DAILY,
                        interval=2,
                        until=dt.date(2011,1,11))
        self.assertEqual(repr(rr),
                        "DTSTART:20110101\n" \
                        "RRULE:FREQ=DAILY;INTERVAL=2;WKST=SU;UNTIL=20110111")
        rr = Recurrence(dtstart=dt.date(2012, 1, 1),
                        freq=YEARLY,
                        bymonth=[1,2],
                        byweekday=range(7),
                        until=dt.date(2012,1,31))
        self.assertEqual(repr(rr),
                        "DTSTART:20120101\n" \
                        "RRULE:FREQ=YEARLY;WKST=SU;UNTIL=20120131;" \
                        "BYDAY=MO,TU,WE,TH,FR,SA,SU;BYMONTH=1,2")
        rr = Recurrence(dtstart=dt.date(2015, 10, 1),
                        freq=MONTHLY,
                        bymonth=range(1,12),
                        byweekday=[(SU(-1))])
        self.assertEqual(repr(rr),
                        "DTSTART:20151001\n" \
                        "RRULE:FREQ=MONTHLY;WKST=SU;BYDAY=-1SU;BYMONTH=1,2,3,4,5,6,7,8,9,10,11")

    def testParse(self):
        rr = Recurrence("DTSTART:20090101\n" \
                        "RRULE:FREQ=WEEKLY;WKST=SU;BYDAY=MO,TU,WE,TH,FR;COUNT=9")
        self.assertEqual(rr.dtstart, dt.date(2009, 1, 1))
        self.assertEqual(rr.count, 9)
        self.assertCountEqual(rr.byweekday, [MO,TU,WE,TH,FR])

    def testParseNoDtstart(self):
        rr = Recurrence("RRULE:FREQ=DAILY;WKST=SU")
        self.assertEqual(rr.freq, DAILY)

    def testRoundtrip(self):
        rrStr = "DTSTART:20151001\n" \
                "RRULE:FREQ=MONTHLY;WKST=SU;BYDAY=-1SU;BYMONTH=1,2,3,4,5,6,7,8,9,10,11"
        self.assertEqual(repr(Recurrence(rrStr)), rrStr)
        rrStr = "DTSTART:20141001\n" \
                "RRULE:FREQ=MONTHLY;WKST=SU;UNTIL=20141001;BYMONTHDAY=1,-1"   # first&last
        self.assertEqual(repr(Recurrence(rrStr)), rrStr)

    def testGetWhen(self):
        rr = Recurrence(dtstart=dt.date(2009, 1, 1),
                        freq=WEEKLY,
                        count=9,
                        byweekday=[MO,WE,FR])
        self.assertEqual(rr._getWhen(1), "Tuesdays, Thursdays and Saturdays")

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------



# ------------------------------------------------------------------------------
# Joyous Groups
# ------------------------------------------------------------------------------
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _
from wagtail.core.models import Page
from wagtail.core.fields import RichTextField
from wagtail.admin.edit_handlers import FieldPanel

# ------------------------------------------------------------------------------
# Events can belong to groups
# by default joyous.GroupPage, but also designed to work with goatherd.GroupPage
# But we don't want a dependency on ls.goatherd, so make this settable
# TODO: consider using https://github.com/wq/django-swappable-models
# ------------------------------------------------------------------------------
def get_group_model_string():
    """
    Get the dotted ``app.Model`` name for the group model as a string.
    Useful for developers that need to refer to the group model, such as in
    foreign keys, but the model itself is not required.
    """
    return getattr(settings, "JOYOUS_GROUP_MODEL", "joyous.GroupPage")


def get_group_model():
    """
    Get the group model from the ``JOYOUS_GROUP_MODEL`` setting.
    Useful for developers that need the group model.
    Defaults to the standard :class:`ls.joyous.models.GroupPage` model
    if no custom model is defined.
    """
    from django.apps import apps
    model_string = get_group_model_string()
    try:
        return apps.get_model(model_string)
    except ValueError:
        raise ImproperlyConfigured("JOYOUS_GROUP_MODEL must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            "JOYOUS_GROUP_MODEL refers to model '%s' that has not been installed" % model_string
        )

# ------------------------------------------------------------------------------
# GroupPage
# ------------------------------------------------------------------------------
class GroupPage(Page):
    class Meta:
        verbose_name = _("group page")
        verbose_name_plural = _("group pages")

    # Define page_ptr so the related_name doesn't clash
    page_ptr = models.OneToOneField(Page, on_delete=models.CASCADE,
                                    related_name="%(app_label)s_%(model_name)s",
                                    parent_link=True)

    subpage_types = ['joyous.SimpleEventPage',
                     'joyous.MultidayEventPage',
                     'joyous.RecurringEventPage',
                     'joyous.MultidayRecurringEventPage']

    content = RichTextField(_("content"), default='', blank=True)
    content.help_text = _("An area of text for whatever you like")

    content_panels = Page.content_panels + [
        FieldPanel('content', classname="full"),
        ]

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
